// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause

#ifndef token_CxxABIConfigure_h
#define token_CxxABIConfigure_h

#define token_HAS_CXXABI_DEMANGLE

#ifdef token_HAS_CXXABI_DEMANGLE
#  include <cxxabi.h>
#endif

#endif // token_CxxABIConfigure_h
