// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
/**
 * @file    vtkOpenXR.h
 * @brief   Defines the OpenXR types and extensions common to all platforms.
 */

#ifndef vtkOpenXR_h
#define vtkOpenXR_h

#include <openxr/openxr.h>

#include "XrExtensions.h"

#define HAND_COUNT 2

#define LEFT_EYE 0
#define RIGHT_EYE 1

#endif
// VTK-HeaderTest-Exclude: vtkOpenXR.h
